#include "BPREngine.h"

BPREngine::BPREngine(
  Matrix ratingMatrix,
  size_t factors,
  size_t iterations,
  size_t seed,
  double userLambda,
  double itemLambda,
  double learningRate,
  size_t numNegativeSamples,
  size_t evalNumNeg,
  int32_t evalSeed,
  double decayRate,
  bool shuffleTrainingSet,
  double initDistributionBound
):  ratingMatrix(ratingMatrix),
    factors(factors),
    iterations(iterations),
    gen_(seed),
    userLambda(userLambda),
    itemLambda(itemLambda),
    _learningRate(learningRate),
    numNegativeSamples(numNegativeSamples),
    evalNumNeg_(evalNumNeg),
    evalSeed_(evalSeed),
    decayRate(decayRate),
    shuffleTrainingSet(shuffleTrainingSet),
    initDistributionBound(initDistributionBound)
{

  this->numUsers = this->ratingMatrix.rows();
  this->numItems = this->ratingMatrix.cols();

  for (size_t i = 0; i < numUsers; i++) {
    for (size_t j = 0; j < numItems; j++) {

      if (ratingMatrix(i, j) < 1.0) {
        continue;
      }
      const size_t uidx = userIndex_.getOrSetIdx(i);
      const size_t pidx = itemIndex_.getOrSetIdx(j);

      data_.push_back(PosPair{uidx, pidx});
    }
  }

  itemMap_.resize(userIndex_.size());
  for (const auto& p : data_) {
    itemMap_[p.userIdx].insert(p.posItemIdx);
  }

  // generate evaluation set
  iterate([& evalSet = evalSet_](PosNegTriplet && triplet) {
    evalSet.push_back(std::move(triplet));
  }, evalNumNeg_, std::mt19937(evalSeed_));

  std::uniform_real_distribution<double> distr(-initDistributionBound, initDistributionBound);
  auto genUnif = [&distr, & gen = gen_](size_t i, size_t j) {
    return distr(gen);
  };

  userFactors = Matrix::random(this->numUsers, this->factors, genUnif);
  itemFactors = Matrix::random(this->numItems, this->factors, genUnif);

}

Matrix BPREngine::optimize() {
  for (size_t i = 0; i < iterations; i++) {

    auto updateOne = [this](const PosNegTriplet& triplet) { update(triplet); };

    iterate(updateOne, numNegativeSamples, gen_);

    evaluate();

    // update learning rate/dataset
    if (decayRate < 1.0) {
      _learningRate *= decayRate;
    }
    if (shuffleTrainingSet) {
      shuffle();
    }
  }

  Matrix X = userFactors * itemFactors.transpose();
  return X;
}

void BPREngine::shuffle() {
  std::shuffle(data_.begin(), data_.end(), gen_);
}

void BPREngine::evaluate() {
  // evaluate on train/test evaluation sets
  auto evalLoss = [this](const PosNegTriplet& triplet) {
    return loss(predictDifference(
      triplet.userIdx, triplet.posItemIdx, triplet.negItemIdx));
  };

  double trainEvalLoss = -1.0;

  if (!evalSet_.empty()) {
    double sum = 0.0;
    for (size_t i = 0; i < evalSet_.size(); i++) {
      sum += evalLoss(evalSet_[i]);
    }
    trainEvalLoss = sum / evalSet_.size();
  }

}

double BPREngine::predictDifference(const size_t userIdx,
                                    const size_t posItemIdx,
                                    const size_t negItemIdx) const {
  // score difference: b_i - b_j + p_u'(q_i - q_j)
  double pred = 0.0;

  for (size_t i = 0; i < factors; ++i) {
    pred += userFactors(userIdx, i) *
            (itemFactors(posItemIdx, i) - itemFactors(negItemIdx, i));
  }
  return pred;
}

double BPREngine::loss(const double scoreDifference) const {
  return log(1.0 + exp(-scoreDifference));
}

double BPREngine::lossDerivative(const double scoreDifference) const {
  return 1.0 / (1.0 + exp(scoreDifference));
}

template <typename FuncT, typename GenT>
void BPREngine::iterate(FuncT func, const size_t numNeg, GenT&& gen) const {
  for (const auto& elem : data_) {
    for (size_t i = 0; i < numNeg; ++i) {
      size_t randNumber = sampleRandomNegative(elem.userIdx, std::forward<GenT>(gen));
      func(PosNegTriplet{
        elem.userIdx,
        elem.posItemIdx,
        randNumber});
    }
  }
}

template <typename GenT>
size_t BPREngine::sampleRandomNegative(const size_t userIdx,
                                       GenT&& gen,
                                       const bool useTestItemMap) const {
  const auto& userPosSet = itemMap_.at(userIdx);
  std::uniform_int_distribution<> dis(0, static_cast<int>(numItems) - 1);
  size_t negIdx;
  do {
    negIdx = dis(gen);
  } while (userPosSet.count(negIdx) > 0);
  return negIdx;
}

void BPREngine::update(const PosNegTriplet& triplet) {
  const size_t uidx = triplet.userIdx;
  const size_t pidx = triplet.posItemIdx;
  const size_t nidx = triplet.negItemIdx;

  const double e = lossDerivative(predictDifference(uidx, pidx, nidx));
  const double lr = _learningRate;

  for (size_t i = 0; i < factors; ++i) {
    const double step =
      lr * (e * (itemFactors(pidx, i) - itemFactors(nidx, i)) -
            userLambda * userFactors(uidx, i));
    userFactors(uidx, i) = userFactors(uidx, i) + step;
  }

  for (size_t i = 0; i < factors; ++i) {
    const double step = lr * (e * userFactors(uidx, i) -
                              itemLambda * itemFactors(pidx, i));
    itemFactors(pidx, i) = itemFactors(pidx, i) + step;
  }

  for (size_t i = 0; i < factors; ++i) {
    const double step = lr * (-e * userFactors(uidx, i) -
                              itemLambda * itemFactors(nidx, i));
    itemFactors(nidx, i) = itemFactors(nidx, i) + step;
  }

}
