#include "LogisticMFEngine.h"

LogisticMFEngine::LogisticMFEngine(
  Matrix ratingMatrix,
  size_t factors, size_t iterations,
  double regularization, double gamma,
  size_t seed
):  ratingMatrix(ratingMatrix),
    regularization(regularization), gamma(gamma),
    factors(factors), iterations(iterations),
    seed(seed)
{
  this->numUsers = this->ratingMatrix.rows();
  this->numItems = this->ratingMatrix.cols();
  this->ones = Matrix::ones(this->numUsers, this->numItems);

  // set seed for random
  std::srand(this->seed);

  this->userVectors = Matrix::random(this->numUsers, this->factors);
  this->itemVectors = Matrix::random(this->numItems, this->factors);
  this->userBiases = Matrix::random(1, this->numUsers)(0);
  this->itemBiases = Matrix::random(1, this->numItems)(0);
}

LogisticMFEngine::LogisticMFEngine(
  Matrix ratingMatrix,
  Matrix userVectors, Matrix itemVectors,
  Vector userBiases, Vector itemBiases,
  double regularization, double gamma,
  size_t iterations, size_t factors
):  ratingMatrix(ratingMatrix),
    userVectors(userVectors), itemVectors(itemVectors),
    userBiases(userBiases), itemBiases(itemBiases),
    regularization(regularization), gamma(gamma),
    factors(factors), iterations(iterations)
{
  this->numUsers = this->ratingMatrix.rows();
  this->numItems = this->ratingMatrix.cols();
  this->ones = Matrix::ones(this->numUsers, this->numItems);
}


void LogisticMFEngine::derive(bool user, Matrix& vecDeriv, Matrix& biasDeriv) {

  if(user)  vecDeriv = this->ratingMatrix * this->itemVectors;
  else      vecDeriv = this->ratingMatrix.transpose() * this->userVectors;

  biasDeriv = this->ratingMatrix.sum(user);

  Matrix A = this->userVectors * this->itemVectors.transpose();
  A.addToRow(this->userBiases);
  A.addToCol(this->itemBiases);
  A.exp();
  A = A / (A + this->ones);
  A = A.multiply(this->ratingMatrix + this->ones);

  if(user) {
    vecDeriv = vecDeriv - ( A * this->itemVectors );
    biasDeriv = biasDeriv - A.sum(user);
    vecDeriv = vecDeriv - ( this->userVectors * this->regularization );
  }
  else {
    vecDeriv = vecDeriv - ( A.transpose() * this->userVectors );
    biasDeriv = biasDeriv - A.sum(user);
    vecDeriv = vecDeriv - ( this->itemVectors * this->regularization );
  }
}

Matrix LogisticMFEngine::train() {

  Matrix userVecDerivSum(this->numUsers, this->factors);
  Matrix itemVecDerivSum(this->numItems, this->factors);
  Matrix userBiasDerivSum(this->numUsers, 1);
  Matrix itemBiasDerivSum(this->numItems, 1);

  for(size_t i = 0; i < this->iterations; i++) {

    Matrix userVecDeriv;
    Matrix userBiasDeriv;
    derive(true, userVecDeriv, userBiasDeriv);

    userVecDerivSum = userVecDerivSum + userVecDeriv.square();
    userBiasDerivSum = userBiasDerivSum + userBiasDeriv.square();

    Matrix userVecStepSize = userVecDerivSum.sqrt().divideBy(this->gamma);
    Matrix userBiasStepSize = userBiasDerivSum.sqrt().divideBy(this->gamma);

    this->userVectors = this->userVectors + (userVecStepSize.multiply(userVecDeriv));
    this->userBiases = this->userBiases + (userBiasStepSize.multiply(userBiasDeriv).transpose())(0);

    // item part
    Matrix itemVecDeriv;
    Matrix itemBiasDeriv;
    derive(false, itemVecDeriv, itemBiasDeriv);

    itemVecDerivSum = itemVecDerivSum + itemVecDeriv.square();
    itemBiasDerivSum = itemBiasDerivSum + itemBiasDeriv.square();

    Matrix itemVecStepSize = itemVecDerivSum.sqrt().divideBy(this->gamma);
    Matrix itemBiasStepSize = itemBiasDerivSum.sqrt().divideBy(this->gamma);

    this->itemVectors = this->itemVectors + (itemVecStepSize.multiply(itemVecDeriv));
    this->itemBiases = this->itemBiases + (itemBiasStepSize.multiply(itemBiasDeriv).transpose())(0);

  }

  return this->userVectors * this->itemVectors.transpose();
}
